//=======================================================================================//
// Q3-BSP.C (C) 2014 Josh PH3NOM Pearson
//=======================================================================================//

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h> 
#include <string.h>

#include <dc/vec3f.h>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>

#include "q3-bsp.h"
#include "rgb.h"
#include "texture.h"
#include "vector.h"

//=======================================================================================//
/* LOCAL VARS */

static GLuint bspmalloc = 0;
static GLuint pvrmalloc = 0;

static byte Q3_BSP_VERBOSE   = 0;
static byte Q3_BSP_RENDER_LM = 0;

static Q3_BSP_HDR         *BSP_HDR       = NULL;
static Q3_BSP_FACE        *BSP_FACE      = NULL;
static Q3_BSP_TEXTURE     *BSP_TEX       = NULL;
static Q3_BSP_VERTEX      *BSP_VERTEX    = NULL;
static Q3_BSP_PLANE       *BSP_PLANES    = NULL;
static Q3_BSP_NODE        *BSP_NODES     = NULL;
static Q3_BSP_BRUSH       *BSP_BRUSHES   = NULL;     
static Q3_BSP_BRUSH_SIDE  *BSP_BRUSH_SIDES = NULL;    
static Q3_BSP_VIS_DATA    *BSP_VIS       = NULL;
static Q3_BSP_LEAF        *BSP_LEAF      = NULL;
static GLuint             *BSP_INDEX     = NULL;
static GLuint             *BSP_LEAF_FACE = NULL;
static GLuint             *BSP_LEAF_BRUSH = NULL;    
static char               *BSP_ENTITIES  = NULL;
static byte               *BSP_FACE_VIS  = NULL;
static byte               *BSP_FACE_CLIP = NULL;

static Q3_BSP_LIGHTMAP_DC *BSP_LIGHTMAP = NULL;
static GLuint LIGHT_TEX[256];
static GlTexture * GL_TEX = NULL;

static Q3_BSP_VERTEX *GL_BSP_VERTEX;
static GLuint        *GL_BSP_VERTICES = 0;
static GLuint        *GL_BSP_TEXID = 0;

static GLuint BSP_LUMP_ALLOC[Q3_BSP_HDR_LUMPS];

#define Q3BSP_DC_LIGHTOFFSET 32 // Brighten Vertex Color Per RGB Component For DC

//===========================================================================================//
/* ENABLE CAPS */

void Q3BSP_Verbose(byte enable)
{
	Q3_BSP_VERBOSE = enable;
}

void Q3BSP_EnableLightMap(byte enable)
{
	Q3_BSP_RENDER_LM = enable;
}

//===========================================================================================//
/* RETURN COUNT FUNCS */

static uint32 Q3BSP_LightMaps()
{
    return BSP_HDR->lump[Q3_LUMP_LIGHT_MAPS].len / sizeof(Q3_BSP_LIGHTMAP);
}

static uint32 Q3BSP_Textures()
{
    return BSP_HDR->lump[Q3_LUMP_TEXTURES].len / sizeof(Q3_BSP_TEXTURE);
}

static uint32 Q3BSP_Faces()
{
    return BSP_HDR->lump[Q3_LUMP_FACES].len / sizeof(Q3_BSP_FACE); 
}

static float Q3BSP_Vertices()
{
     return BSP_HDR->lump[Q3_LUMP_VERTICES].len / (float)sizeof(Q3_BSP_VERTEX); 
}

static uint32 Q3BSP_Entities()
{
    return BSP_HDR->lump[Q3_LUMP_ENTITIES].len;
}

static uint32 Q3BSP_Planes()
{
     return BSP_HDR->lump[Q3_LUMP_PLANES].len / sizeof(Q3_BSP_PLANE); 
}

static uint32 Q3BSP_Nodes()
{
    return BSP_HDR->lump[Q3_LUMP_NODES].len / sizeof(Q3_BSP_NODE); 
}

static uint32 Q3BSP_Leaves()
{
    return BSP_HDR->lump[Q3_LUMP_LEAFS].len / sizeof(Q3_BSP_LEAF); 
}

static uint32 Q3BSP_LeafFaces()
{
    return BSP_HDR->lump[Q3_LUMP_LEAF_FACES].len; 
}

//===========================================================================================//
/* HELPER FUNCS */

static void * Q3BSP_Malloc(uint32 s)
{
	bspmalloc += s;
	return malloc(s);
}

static void Q3BSP_Free(uint32 lump)
{
	printf("Q3BSP: Have %i bytes before deallocation\n", bspmalloc);
	
	switch(lump)
	{
	    case Q3_LUMP_VERTICES:
			 printf("Q3BSP: Free Q3_LUMP_VERTICES\n");
		     free(BSP_VERTEX);
			 BSP_VERTEX = NULL;
			 bspmalloc -= BSP_LUMP_ALLOC[lump];
			 break;

	    case Q3_LUMP_FACE_INDICES:
			 printf("Q3BSP: Free Q3_LUMP_FACE_INDICES\n");
		     free(BSP_INDEX);
			 BSP_INDEX = NULL;
			 bspmalloc -= BSP_LUMP_ALLOC[lump];
			 break;

	    case Q3_LUMP_LEAF_FACES:
			 printf("Q3BSP: Free Q3_LUMP_LEAF_FACES\n");
		     free(BSP_LEAF_FACE);
			 BSP_LEAF_FACE = NULL;
			 bspmalloc -= BSP_LUMP_ALLOC[lump];
			 break;
	}

	printf("Q3BSP: Have %i bytes after deallocation\n", bspmalloc);
}

static FILE* Q3BSP_FOPEN(char *fname)
{
	FILE *f = NULL;
	
	f = fopen(fname, "rb");

	if(f == NULL)
	{
	    printf("Error Opening %s\n", fname);
		while(1);
	}

	return f;
}

static inline void Q3BSP_VectorSwizzle3f(vec3f_t *v)
{
	float z = v->y;
	v->x = -v->x;
	v->y = v->z;
	v->z = z;
}

static inline void Q3BSP_VectorSwizzle3i(int *i)
{
	int z = i[1];
	i[0] = -i[0];
	i[1] = i[2];
	i[2] = z;
}

//===========================================================================================//
/* LOAD FUNCS */

char bsp_file_name[128];

static byte Q3BSP_LoadHeader(char *fname)
{
	FILE *bsp = Q3BSP_FOPEN(fname);

	if(BSP_HDR != NULL)
	{
		free(BSP_HDR);
	    bspmalloc -= sizeof(Q3_BSP_HDR);
	}
	BSP_HDR = Q3BSP_Malloc(sizeof(Q3_BSP_HDR));

	fseek(bsp, 0, SEEK_SET);
	fread(BSP_HDR, 1, sizeof(Q3_BSP_HDR), bsp);
	fclose(bsp);

	if(BSP_HDR->version != Q3_BSP_HDR_VERSION)
		return 0;
	if(BSP_HDR->id[0] != 'I' || BSP_HDR->id[1] != 'B' || BSP_HDR->id[2] != 'S' || BSP_HDR->id[3] != 'P')
	    return 0;

	return 1;
}

static void Q3BSP_LoadLump(char *fname, byte lump)
{
	FILE *f = Q3BSP_FOPEN(fname);

	fseek(f, BSP_HDR->lump[lump].oft, SEEK_SET);
	
	BSP_LUMP_ALLOC[lump] = BSP_HDR->lump[lump].len;

	GLuint i;

	switch(lump)
	{
	    case Q3_LUMP_BRUSHES:
			 printf("Q3_LUMP_BRUSHES: Loading %i bytes\n", BSP_HDR->lump[lump].len);
		     BSP_BRUSHES = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_BRUSHES, 1, BSP_HDR->lump[lump].len, f);
			 break;

		case Q3_LUMP_BRUSH_SIDES:
			 printf("Q3_LUMP_BRUSH_SIDES: Loading %i bytes\n", BSP_HDR->lump[lump].len);
		     BSP_BRUSH_SIDES = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_BRUSH_SIDES, 1, BSP_HDR->lump[lump].len, f);
			 break;

		case Q3_LUMP_LEAF_BRUSHES:
			 printf("Q3_LUMP_LEAF_BRUSHES: Loading %i bytes\n", BSP_HDR->lump[lump].len);
		     BSP_LEAF_BRUSH = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_LEAF_BRUSH, 1, BSP_HDR->lump[lump].len, f);
			 break;

	    case Q3_LUMP_ENTITIES:
			 BSP_ENTITIES = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_ENTITIES, 1, BSP_HDR->lump[lump].len, f);
			 break;

		case Q3_LUMP_PLANES:
			 BSP_PLANES = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_PLANES, 1, BSP_HDR->lump[lump].len, f);

			 for(i = 0; i < Q3BSP_Planes(); i++)
			 {
				 Q3BSP_VectorSwizzle3f(&BSP_PLANES[i].normal.x);
			     BSP_PLANES[i].intercept = -BSP_PLANES[i].intercept;
			 }
			 break;

		case Q3_LUMP_NODES:
			 BSP_NODES = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_NODES, 1, BSP_HDR->lump[lump].len, f);
			 break;

		case Q3_LUMP_LEAFS:
			 BSP_LEAF = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_LEAF, 1, BSP_HDR->lump[lump].len, f);

			 for(i = 0; i < Q3BSP_Leaves(); i++)
			 {
				 Q3BSP_VectorSwizzle3i(&BSP_LEAF[i].mins[0]);
			     Q3BSP_VectorSwizzle3i(&BSP_LEAF[i].maxs[0]);				 
			 }

			 break;

	    case Q3_LUMP_FACES:
			 BSP_FACE = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_FACE, 1, BSP_HDR->lump[lump].len, f);
             break;

		case Q3_LUMP_LEAF_FACES:
			 BSP_LEAF_FACE = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_LEAF_FACE, 1, BSP_HDR->lump[lump].len, f);
			 break;

	    case Q3_LUMP_VERTICES:
			 BSP_VERTEX = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_VERTEX, 1, BSP_HDR->lump[lump].len, f);

			 for(i = 0; i < Q3BSP_Vertices(); i++)
			 {
				 BSP_VERTEX[i].color = ARGB32_ADD(BSP_VERTEX[i].color, Q3BSP_DC_LIGHTOFFSET);

				 Q3BSP_VectorSwizzle3f(&BSP_VERTEX[i].pos.x);
				 Q3BSP_VectorSwizzle3f(&BSP_VERTEX[i].normal.x);
			 }
			 break;

		case Q3_LUMP_FACE_INDICES:
			 BSP_INDEX = Q3BSP_Malloc(BSP_HDR->lump[lump].len);
			 fread(BSP_INDEX, 1, BSP_HDR->lump[lump].len, f);
			 break;

		case Q3_LUMP_VIS_DATA:
			 BSP_VIS = Q3BSP_Malloc(sizeof(Q3_BSP_VIS_DATA));
			 fread(BSP_VIS, 2, sizeof(int), f);

			 printf("Q3BSP: Vis Data: %i Clusters || %i Bytes/Cluster\n", BSP_VIS->clusters, BSP_VIS->bytesPerCluster);

			 BSP_VIS->bitset = Q3BSP_Malloc(BSP_VIS->clusters * BSP_VIS->bytesPerCluster);
			 fread(BSP_VIS->bitset, BSP_VIS->bytesPerCluster, BSP_VIS->clusters, f);

			 BSP_FACE_VIS = Q3BSP_Malloc(Q3BSP_Faces());
			 BSP_FACE_CLIP = Q3BSP_Malloc(Q3BSP_Faces());
			 Q3BSP_ClearVisData();

			 break;
	}

	fclose(f);
}

void Q3BSP_LoadTextures(char *fname)
{
    FILE *f = Q3BSP_FOPEN(fname);

    fseek(f, BSP_HDR->lump[Q3_LUMP_TEXTURES].oft, SEEK_SET);
	
    BSP_TEX = Q3BSP_Malloc(BSP_HDR->lump[Q3_LUMP_TEXTURES].len);

    fread(BSP_TEX, 1, BSP_HDR->lump[Q3_LUMP_TEXTURES].len, f);
    fclose(f);

    int tex = Q3BSP_Textures();

    if(Q3_BSP_VERBOSE)
        printf("BSP: Loading %i Textures...\n", tex);

    GL_TEX = Q3BSP_Malloc(tex * sizeof(GlTexture));

    char texname[64+16];
    char ptexname[64+16];
    int i;
    for(i=0; i<tex; i++)
    {
        if(!memcmp("noshader", BSP_TEX[i].name, 8)) // noshader = no texture to decode //
        {
            GL_TEX[i].data = NULL;
                continue;
        }
        
        sprintf(ptexname, "/cd/%s.pvr", BSP_TEX[i].name);        
        sprintf(texname, "/cd/%s.jpg", BSP_TEX[i].name);            
    
        GL_TEX[i].ID = glTextureLoadPVR(ptexname, 0);
        
        if(GL_TEX[i].ID) 
        {
             printf("PVRTEX %s\n", ptexname);
             
             continue;
        }      
        
        printf("PVR2GLTEX Failure %s\n",texname);

        if(!glTextureLoadJPG(texname, &GL_TEX[i])) // decode jpeg texture and bind to gl //
        {
             printf("JPEG2GLTEX Failure %s\n",texname);
             
             GL_TEX[i].ID = 1;             
        }
        else
        {
            pvrmalloc += GL_TEX[i].TexW * GL_TEX[i].TexH * 2;
        
            if(Q3_BSP_VERBOSE)
            {
                printf("BSP: Texture[%i] %s\n     Flags: %i || Contents: %i\n", i, texname, BSP_TEX[i].flags, BSP_TEX[i].contents );
                printf("     Resolution: Img = %ix%i || Tex = %ix%i\n", GL_TEX[i].ImgW, GL_TEX[i].ImgH,
			                                                        GL_TEX[i].TexW, GL_TEX[i].TexH );

                if(GL_TEX[i].TexW != GL_TEX[i].ImgW || GL_TEX[i].TexH != GL_TEX[i].ImgH)
	                printf("Warning: Texture U/V is SCEWED!\n"); // Need to compensate for u/v offset - scalar float i/t on vertex //
            }
        }

    }
}

void Q3BSP_LoadLightMaps(char *fname)
{
    FILE * f =  Q3BSP_FOPEN(fname);

    uint32 tex = Q3BSP_LightMaps();
    int pm = pvrmalloc;
    if(Q3_BSP_VERBOSE)
    {
        printf("BSP: Loading %i Light Maps...\n", tex);
        printf("PVR MEM USED: %i\n", pvrmalloc);
    }
    Q3_BSP_LIGHTMAP lm;
    unsigned short *p16 = BSP_LIGHTMAP;
    uint32 i,o;

    fseek(f, BSP_HDR->lump[Q3_LUMP_LIGHT_MAPS].oft, SEEK_SET);
    
    unsigned short * TEX0 = malloc(glKosMipMapTexSize(128, 128));

    for( o = 0; o < tex; o++ )
    {
        fread(lm, 1, sizeof(Q3_BSP_LIGHTMAP), f);
		
        unsigned short *p16 = TEX0;
        
        for(i = 0; i < 128 * 128 * 3; i += 3)
            *p16++ = RGB24_2_RGB16(&lm[i]);

        //gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, 128, 128, GL_RGB, GL_UNSIGNED_SHORT_5_6_5, TEX0);
        
        glGenTextures(1, &LIGHT_TEX[o]);
        glBindTexture(GL_TEXTURE_2D, LIGHT_TEX[o]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_FILTER, GL_LINEAR);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	    glTexImage2D( GL_TEXTURE_2D,
                      0,
                      GL_RGB,
                      128, 128, 0,
                      GL_RGB,
                      GL_UNSIGNED_SHORT_5_6_5,
                      TEX0 /*+ 128*128*/ );
	   
        pvrmalloc += 128*128*2;//glKosMipMapTexSize(128, 128);
    }
    
    if(Q3_BSP_VERBOSE)
    {
        printf("BSP: Done Loading %i Light Maps...\n", tex);
        printf("PVR MEM USED by LightMaps: %i\n", pvrmalloc - pm);
    }
    
    free(TEX0);
    
    fclose(f);
}

bool Q3BSP_Load(char *fname, vec3f_t *startPos)
{
    printf("\n\nQ3BSP: Loading %s\n", fname);

    if(!Q3BSP_LoadHeader(fname))
    {
        printf("Q3BSP: ERROR BSP Header Invalid Version\n");
        return 0;
    }
    else if(Q3_BSP_VERBOSE)
        printf("Q3BSP: Header Loaded and Verified. Loading Lumps...\n\n");
	
	GLubyte i;
	for(i = 0; i < Q3_BSP_HDR_LUMPS; i++)
	    BSP_LUMP_ALLOC[i] = 0;

	printf("Q3BSP: Loading Lump: Vertices, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadLump(fname, Q3_LUMP_VERTICES);
	printf("Loading Lump: FACE_INDICES, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadLump(fname, Q3_LUMP_FACE_INDICES);
	printf("Loading Lump: LUMP_FACES, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadLump(fname, Q3_LUMP_FACES);

	printf("Loading Lump: Textures, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadTextures(fname);
	printf("Loading Lump: LightMaps, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadLightMaps(fname);

	strcpy(bsp_file_name, fname);

    if(Q3_BSP_VERBOSE)
    {
        printf("\nQ3BSP:  RAM: %i\n", bspmalloc);
        printf("Q3BSP: VRAM: %i\n\n", pvrmalloc);
    }

    printf("Q3BSP: Loading completed\n\n");

    return 1;
}

void Q3BSP_LoadCollisionLumps(char *fname)
{
	printf("Loading Lump: LEAFS, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadLump(fname, Q3_LUMP_LEAFS);

	printf("Loading Lump: PLANES, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadLump(fname, Q3_LUMP_PLANES);
	printf("Loading Lump: NODES, bspmalloc: %i\n", bspmalloc);
    Q3BSP_LoadLump(fname, Q3_LUMP_NODES);

	printf("Loading Lump: BRUSHES, bspmalloc: %i\n", bspmalloc);
	Q3BSP_LoadLump(fname, Q3_LUMP_BRUSHES);
	printf("Loading Lump: BRUSH_SIDES, bspmalloc: %i\n", bspmalloc);
	Q3BSP_LoadLump(fname, Q3_LUMP_BRUSH_SIDES);
	printf("Loading Lump: LEAF_BRUSHES, bspmalloc: %i\n", bspmalloc);
	Q3BSP_LoadLump(fname, Q3_LUMP_LEAF_BRUSHES);
}

//===========================================================================================//
/* RENDER FUNCS */

static Q3_BSP_VERTEX **BSP_FACE_VERTEX;

static inline void Q3BSP_VertexCopy(Q3_BSP_VERTEX *src, Q3_BSP_VERTEX *dst)
{
	*dst = *src;
}

static int **BSP_TEX_INDICES;
static int   *BSP_TEX_INDEX_COUNT;
static int   *BSP_LM_INDEX_COUNT;

int Q3BSP_TextureIndexCount( int texIndex )
{
    Q3_BSP_FACE * face = BSP_FACE;
    
    int count = 0, i;
    
    for(i = 0; i < Q3BSP_Faces(); i++)
    {
        if( face->textureID == texIndex && (face->type == 1 || face->type == 3) )
            count += face->faceIndices;

        ++face;
    }
    
    return count;
}

int Q3BSP_LightMapIndexCount( int lmIndex )
{
    Q3_BSP_FACE * face = BSP_FACE;
    
    int count = 0, i;
    
    for(i = 0; i < Q3BSP_Faces(); i++)
    {
        if( face->lightmapID == lmIndex && (face->type == 1 || face->type == 3) )
            count += face->faceIndices;

        ++face;
    }
    
    return count;
}

/////////////////////////////////////////////


static Q3_BSP_VERTEX mv[8000];

static GLuint Q3BSP_LMArrayCount()
{
    GLuint i, f, faces = Q3BSP_Faces(), index, t, l, count = 0;

    Q3_BSP_FACE * face = BSP_FACE;

    for( l = 0; l < Q3BSP_LightMaps(); l++ )
    {
        for( t = 0; t < Q3BSP_Textures(); t++ )
        {       
            if( !memcmp("noshader", BSP_TEX[t].name, 8) )
                continue;
        
            int verts = 0;
        
            for(i = 0; i < faces; i++)
            {
                if( face->lightmapID == l && face->textureID == t && ( face->type==1 || face->type==3 ) )
                    ++verts;
                ++face;
            }

            face = BSP_FACE;

            if(verts)
                ++count;
        }
    }
    return count;
}

typedef struct
{
    GLuint lmID;
    GLuint tID;
    GLint flags;
    GLuint count;
    Q3_BSP_VERTEX * data;
} Q3BSP_RenderNode;

static Q3BSP_RenderNode * renderNode;
static GLuint renderNodes;

void Q3BSP_CacheArraysMultiTex()
{
    GLuint i, f, faces = Q3BSP_Faces(), index, t, l;

    Q3_BSP_FACE * face = BSP_FACE;
    
    renderNodes = Q3BSP_LMArrayCount();
    
    printf("Q3BSP: %i unique arrays\n", renderNodes);
    
	printf("Q3BSP: Allocating %i bytes for RenderNodes\n", sizeof(Q3BSP_RenderNode) * renderNodes );
    renderNode = Q3BSP_Malloc( sizeof(Q3BSP_RenderNode) * renderNodes );

	printf("Q3BSP: %i bytes in RAM\n", bspmalloc );

    GLuint ap = 0;

  for( l = 0; l < Q3BSP_LightMaps(); l++ )
  {
    for( t = 0; t < Q3BSP_Textures(); t++ )
    {       
        if( !memcmp("noshader", BSP_TEX[t].name, 8) )
            continue;
        
        int verts = 0;
        
        for(i = 0; i < faces; i++)
        {
            if( face->lightmapID == l && face->textureID == t && ( face->type==1 || face->type==3 ) )
            {
	        for(f = 0; f < face->faceIndices; f++)
                {
		    index = face->vertexIndex + BSP_INDEX[face->faceIndex + f];
                    
                    Q3BSP_VertexCopy( &BSP_VERTEX[index], &mv[verts++]);

					if(verts >= 7999)
					{
						printf("ERROR! WARNING: BUFFER OVERFLOW!\n");
					    while(1);
					}
                }	
            }
            ++face;
        }

        face = BSP_FACE;

        if(verts)
        {
            renderNode[ap].count = verts;
            renderNode[ap].lmID = LIGHT_TEX[l];
            renderNode[ap].tID = GL_TEX[t].ID;
            renderNode[ap].flags = BSP_TEX[t].flags;
			printf("Q3BSP: Allocating %i bytes for RenderNode %i\n", sizeof(Q3_BSP_VERTEX) * verts, ap );
            renderNode[ap].data = Q3BSP_Malloc( sizeof(Q3_BSP_VERTEX) * verts );
			printf("Q3BSP: %i bytes in RAM\n", bspmalloc );
            memcpy( renderNode[ap].data, mv, sizeof(Q3_BSP_VERTEX) * verts);
            ++ap;
        }
    }
  }

  printf("Q3BSP: GL Render Nodes Built. Freeing BSP data...\n");
  Q3BSP_Free(Q3_LUMP_LEAF_FACES);
  Q3BSP_Free(Q3_LUMP_VERTICES);
  Q3BSP_Free(Q3_LUMP_FACE_INDICES);

  printf("\nQ3BSP: Done deallocating, now loading Collision Lumps...\n");
  Q3BSP_LoadCollisionLumps(bsp_file_name);

  printf("Q3BSP: Loading Completed\n");
}

void Q3BSP_RenderArraysMultiTex()
{
    GLuint i;

    for( i = 0; i < renderNodes; i++ )
    {
        if(renderNode[i].flags)
            glEnable(GL_BLEND);
 
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, renderNode[i].tID);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_FILTER, GL_LINEAR); 
        
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, renderNode[i].lmID);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_FILTER, GL_LINEAR);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);     
        glActiveTextureARB(GL_TEXTURE0_ARB);

        glBlendFunc(GL_SRC_ALPHA, GL_DST_ALPHA);  
        
        glClientActiveTextureARB(GL_TEXTURE0_ARB); 
        glTexCoordPointer(2, GL_FLOAT, sizeof(Q3_BSP_VERTEX), &renderNode[i].data[0].uv.x);

        glClientActiveTextureARB(GL_TEXTURE1_ARB); 
        glTexCoordPointer(2, GL_FLOAT, sizeof(Q3_BSP_VERTEX), &renderNode[i].data[0].uv.z); 
        glClientActiveTextureARB(GL_TEXTURE0_ARB);   
        
        glColorPointer(1, GL_UNSIGNED_INT, sizeof(Q3_BSP_VERTEX), &renderNode[i].data[0].color);
        
        glVertexPointer(3, GL_FLOAT, sizeof(Q3_BSP_VERTEX), &renderNode[i].data[0].pos);
        
        glDrawArrays(GL_TRIANGLES, 0, renderNode[i].count);   
     
        glDisable(GL_BLEND);     
    } 
}   

