/*
** texture.c (C) Josh PH3NOM Pearson
**
** Read a PVR texture from a file and bind to a GL texture
** Demonstrates glTexImage2D using the Dreamcast OpenGL API
*/

#include <stdio.h>
#include <string.h>

#include "texture.h"


GLuint glTextureLoadPVR(char *fname, unsigned char UseMipMap) {
#define PVR_HDR_SIZE 0x20
    FILE *tex = NULL;
    unsigned char *texBuf;
    GLuint texID, texSize;

    tex = fopen(fname, "rb");

    if(tex == NULL) {
        printf("FILE READ ERROR: %s\n", fname);

        return 0;
    }

    fseek(tex, 0, SEEK_END);
    texSize = ftell(tex);

    texBuf = malloc(texSize);
    fseek(tex, 0, SEEK_SET);
    fread(texBuf, 1, texSize, tex);
    fclose(tex);

    int texW = texBuf[PVR_HDR_SIZE - 4] | texBuf[PVR_HDR_SIZE - 3] << 8;
    int texH = texBuf[PVR_HDR_SIZE - 2] | texBuf[PVR_HDR_SIZE - 1] << 8;
    int texFormat, texColor;

    printf("PVR TEXTURE RESOLUTION: %ix%i\n", texW, texH);
    
    switch((unsigned int)texBuf[PVR_HDR_SIZE - 8]) {
        case 0x00:
            texColor = PVR_TXRFMT_ARGB1555;
            break; //(bilevel translucent alpha 0,255)

        case 0x01:
            texColor = PVR_TXRFMT_RGB565;
            break; //(non translucent RGB565 )

        case 0x02:
            texColor = PVR_TXRFMT_ARGB4444;
            break; //(translucent alpha 0-255)

        case 0x03:
            texColor = PVR_TXRFMT_YUV422;
            break; //(non translucent UYVY )

        case 0x04:
            texColor = PVR_TXRFMT_BUMP;
            break; //(special bump-mapping format)

        case 0x05:
            texColor = PVR_TXRFMT_PAL4BPP;
            break; //(4-bit palleted texture)

        case 0x06:
            texColor = PVR_TXRFMT_PAL8BPP;
            break; //(8-bit palleted texture)

        default:
            texColor = PVR_TXRFMT_RGB565;
            break;
    }

    switch((unsigned int)texBuf[PVR_HDR_SIZE - 7]) {
        case 0x01:
            texFormat = PVR_TXRFMT_TWIDDLED;
            break;//SQUARE TWIDDLED

        case 0x03:
            texFormat = PVR_TXRFMT_VQ_ENABLE;
            break;//VQ TWIDDLED

        case 0x09:
            texFormat = PVR_TXRFMT_NONTWIDDLED;
            break;//RECTANGLE

        case 0x0B:
            texFormat = PVR_TXRFMT_STRIDE | PVR_TXRFMT_NONTWIDDLED;
            break;//RECTANGULAR STRIDE

        case 0x0D:
            texFormat = PVR_TXRFMT_TWIDDLED;
            break;//RECTANGULAR TWIDDLED

        case 0x10:
            texFormat = PVR_TXRFMT_VQ_ENABLE | PVR_TXRFMT_NONTWIDDLED;
            break;//SMALL VQ

        default:
            texFormat = PVR_TXRFMT_NONE;
            break;
    }

    glGenTextures(1, &texID);
    glBindTexture(GL_TEXTURE_2D, texID);
    
    if(texFormat & PVR_TXRFMT_VQ_ENABLE)
        glCompressedTexImage2D(GL_TEXTURE_2D,
                               UseMipMap,
 	                       texFormat | texColor,
 	                       texW,
 	                       texH,
 	                       0,
 	                       texSize - PVR_HDR_SIZE,
 	                       texBuf + PVR_HDR_SIZE);
    else
        glTexImage2D(GL_TEXTURE_2D, UseMipMap, GL_RGB,
                     texW, texH, 0,
                     GL_RGB, texFormat | texColor, texBuf + PVR_HDR_SIZE);

    free(texBuf);
    
    return texID;
}

/* Load a JPEG into a gl texture */
GLuint glTextureLoadJPG(char *fname, GlTexture *texture)
{
	FILE *infile = NULL;
	int		i;
	uint16		* ourbuffer;
	uint16		* temp_tex;
    int scale = 1;

	infile = fopen(fname, "rb");
	if(infile == NULL)
		return 0;

	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	JSAMPARRAY	buffer;			/* Output row buffer */
	int		row_stride;		/* physical row width in output buffer */
    
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, (FILE*)infile);
	(void)jpeg_read_header(&cinfo, TRUE);
	jpeg_calc_output_dimensions(&cinfo);

	texture->ImgW = cinfo.image_width;
	texture->ImgH = cinfo.image_height;
    texture->TexW = texture->TexH = 0x8;
    while(texture->TexW < texture->ImgW)(texture->TexW <<= 1);
    while(texture->TexH < texture->ImgH)(texture->TexH <<= 1);

	texture->data = (void *)( temp_tex = (uint16 *)malloc(texture->ImgW * texture->ImgH * 2) );	
	ourbuffer = (uint16 *)malloc(texture->ImgW * 2);

	cinfo.scale_denom = scale;
	(void)jpeg_start_decompress(&cinfo);
	row_stride = cinfo.output_width * cinfo.output_components;
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	while (cinfo.output_scanline < cinfo.output_height)
	{
		(void)jpeg_read_scanlines(&cinfo, buffer, 1);

		for(i=0; i<row_stride/3; i++)
			ourbuffer[i] = ((buffer[0][i*3]>>3)<<11) + ((buffer[0][i*3+1]>>2)<<5) + (buffer[0][i*3+2]>>3);

		memcpy(temp_tex + cinfo.image_width*(cinfo.output_scanline-1), ourbuffer, 2*row_stride/3); 
	}

	(void)jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

	free(ourbuffer);
	fclose(infile);

    glGenTextures(1, &texture->ID);
    glBindTexture(GL_TEXTURE_2D, texture->ID);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexImage2D(GL_TEXTURE_2D,
                 0,
                 GL_RGB,
                 texture->TexW, texture->TexH,
                 0,
                 GL_RGB,
                 PVR_TXRFMT_NONTWIDDLED | PVR_TXRFMT_RGB565,
                 texture->data);

	free(texture->data); 

	return texture->ID;
}